/* Copyright (C) 2003, 2005 Peter Selinger.
   This file is part of farsiweb. It is free software and it is covered
   by the GNU General Public License. See the file COPYING for details. */

/* ---------------------------------------------------------------------- */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "codetable.h"

/* name, version, and copyright strings */
#define NAME      "farsiweb"
#define VERSION   "0.1"
#define COPYRIGHT "Copyright (C) 2003, 2005 Peter Selinger" 

#define bufsize 255 /* must be long enough to hold each trans[n].code */

/* HTML codes */
#define RTL_BEGIN "<!-- BEGIN FARSI --><bde dir=\"rtl\">"
#define RTL_END   "</bde><!-- END FARSI -->"
#define LTR_BEGIN "<!-- BEGIN LATIN --><bde dir=\"ltr\">"
#define LTR_END   "</bde><!-- END LATIN -->"

/* ---------------------------------------------------------------------- */

/* main procedure: translate fin to fout */
void translate(FILE *fin, FILE *fout)
{
  int ch = 0;         /* remains EOF on end of file */
  struct trans_s *p;
  char buf[bufsize];
  int n = 0;          /* number of characters in buffer */
  int farsi_mode = 0; /* odd = we are in Farsi mode */
  int literal_mode = 0; /* in literal mode, just scan for escapes and end-literals */
  int word_boundary = 1;
  int lon, l, shift;
  struct trans_s *match;

  while (1) {
    /* fill the buffer */
    while (n<bufsize) {
      if (ch != EOF) {
        ch=fgetc(fin);
      }
      buf[n] = ch;
      n++;
    }
    /* empty the buffer - find longest match. */
    lon = 0;
    match = NULL;
    for (p=trans; p->uni; p++) {
      if (p->code == NULL) {
        continue;
      }
      l = strlen(p->code);
      if (bufsize < l || bufsize < 2) {
        fprintf(stderr, "Error: bufsize constant incorrectly defined; fix program and recompile\n");
        exit(1);
      }
      if (l<=lon) {
	continue;
      }
      if (literal_mode && p->typ != literal_mode && p->typ != C_ESC) {
	/* in literal mode only consider C_ELIT and C_ESC tokens */
	continue;
      }
      if (l>lon) {
        if (strncmp(p->code, buf, l)==0) { /* found a match */
          lon = l;
          match = p;
        }
      }
    }
    if (!match) {
      fputc(buf[0], fout);
      shift = 1;
      word_boundary = 1;
    } else if (literal_mode && match->typ == literal_mode) {
      literal_mode = 0;
      fputs(match->uni, fout);
      shift = lon;
    } else if (literal_mode && match->typ == C_ESC) {
      fputc(buf[1], fout);
      word_boundary = 1;
      shift = 2;
    } else if (literal_mode) {
      fputc(buf[0], fout);
      shift = 1;
    } else if (match->typ == C_BEGIN) {
      farsi_mode++;
      fputs((farsi_mode & 1) ? RTL_BEGIN : LTR_BEGIN, fout);
      word_boundary = 1;
      shift = lon;
    } else if (match->typ == C_END) {
      fputs((farsi_mode & 1) ? RTL_END : LTR_END, fout);
      farsi_mode--;
      word_boundary = 1;
      shift = lon;
    } else if ((farsi_mode & 1) == 0) {
      fputc(buf[0], fout);
      shift = 1;
    } else if (IS_BLIT(match->typ)) {
      literal_mode = TO_ELIT(match->typ);
      fputs(match->uni, fout);
      shift = lon;
    } else if (match->typ == C_ESC) {
      fputc(buf[1], fout);
      word_boundary = 1;
      shift = 2;
    } else {
      if (word_boundary && match->alt != 0) {
        fputs(match->alt, fout);
      } else {
        fputs(match->uni, fout);
      }
      shift = lon;
      if (match->typ == C_LET || match->typ == C_MOD || match->typ == C_NULL) {
        word_boundary = 0;
      } else {
        word_boundary = 1;
      }
    }
    /* shift the buffer */
    memmove(buf, buf+shift, (n-shift)*sizeof(buf[0]));
    n -= shift;
    if (buf[0] == EOF) {
      break;
    }
  }
  return;
}

/* ---------------------------------------------------------------------- */

/* make an HTML table of unicode characters from n0 to n1 */
void unicodetable(FILE *fout, int n0, int n1) {
  int i;
  int col=0;
  int row=0;

  for (i=n0; i<n1; i++) {
    if (row == 0 && col == 0) {
      fprintf(fout, "<table border=0>\n");
    }
    if (col == 0) {
      fprintf(fout, "<tr>\n");
    }
    fprintf(fout, "<td bgcolor=#ffffc0>");
    fprintf(fout, "%d", i);
    fprintf(fout, "</td>\n");
    fprintf(fout, "<td bgcolor=#ffc0ff>");
    fprintf(fout, "%04x", i);
    fprintf(fout, "</td>\n");
    fprintf(fout, "<td bgcolor=#c0ffff><font size=+2>");
    fprintf(fout, "&#%d;", i);
    fprintf(fout, "</font></td>\n");
    fprintf(fout, "<td bgcolor=#c0ffff><font size=+2>");
    fprintf(fout, "&#1600;&#%d;", i);
    fprintf(fout, "</font></td>\n");
    fprintf(fout, "<td bgcolor=#c0ffff><font size=+2>");
    fprintf(fout, "&#%d;&#1600;", i);
    fprintf(fout, "</font></td>\n");
    fprintf(fout, "<td bgcolor=#c0ffff><font size=+2>");
    fprintf(fout, "&#1600;&#%d;&#1600;", i);
    fprintf(fout, "</font></td>\n");
    col++;
    if (col == 4) {
      fprintf(fout, "</tr>\n");
      col=0;
      row++;
    }
    if (row == 16) {
      fprintf(fout, "</table>\n");
      row=0;
      col=0;
    }
  }
  if (col != 0) {
    fprintf(fout, "</tr>\n");
  }
  if (row != 0 || col != 0) {
    fprintf(fout, "</table>\n");
  }    
}

/* ---------------------------------------------------------------------- */

/* make a table of transliteration codes used in this program */
void codetable(FILE *fout) {
  struct trans_s *p;
  int row=0;

  fprintf(fout, "<title>Code table</title>\n");
  fprintf(fout, "<h1>Code table</h1>\n");
  fprintf(fout, "<table border=0 cellspacing=3>\n");
  fprintf(fout, "<tr><td>Legend:</td></tr>\n");
  fprintf(fout, "<tr><td bgcolor=#ffffc0>yellow: code</td></tr>\n");
  fprintf(fout, "<tr><td bgcolor=#ffc0ff>magenta: translation</td></tr>\n");
  fprintf(fout, "<tr><td bgcolor=#c0ffff>cyan: connected forms</td></tr>\n");
  fprintf(fout, "<tr><td bgcolor=#ffe0e0>pink: alternate translation at start of word</td></tr>\n");
  fprintf(fout, "<tr><td bgcolor=#e0e0ff>blue: name of symbol</td></tr>\n");
  fprintf(fout, "</table>\n");

  fprintf(fout, "<table border=0 cellspacing=3><tr>\n");
  for (p=trans; p->uni; p++) {
    if (p->code != NULL) {
      if (row==0) {
        fprintf(fout, "<td valign=top><table border=0>\n");
      }
      fprintf(fout, "<tr>\n");
      fprintf(fout, "<td bgcolor=#ffffc0 align=center>%s</td>", p->code);
      fprintf(fout, "<td bgcolor=#ffc0ff align=center><font size=+2><nobr>%s</nobr></font></td>", p->uni);
      if (p->typ == C_LET) {
        fprintf(fout, "<td bgcolor=#c0ffff align=center><font size=+2><nobr>&#1600;%s</nobr></font></td>", p->uni);
        fprintf(fout, "<td bgcolor=#c0ffff align=center><font size=+2><nobr>%s&#1600;</nobr></font></td>", p->uni);
        fprintf(fout, "<td bgcolor=#c0ffff align=center><font size=+2><nobr>&#1600;%s&#1600;</nobr></font></td>", p->uni);
      } else if (p->typ == C_MOD) {
        fprintf(fout, "<td bgcolor=#c0ffff align=center><font size=+2><nobr>%s&#1600;</nobr></font></td>", p->uni);
        fprintf(fout, "<td bgcolor=#c0ffff colspan=2>&nbsp;</td>");
      } else {
        fprintf(fout, "<td bgcolor=#c0ffff colspan=3>&nbsp;</td>");
      }
      if (p->alt) {
	fprintf(fout, "<td bgcolor=#ffe0e0 align=center><font size=+2><nobr>%s</nobr></font></td>", p->alt);
      } else {
	fprintf(fout, "<td bgcolor=#ffe0e0>&nbsp;</td>");
      }	
      fprintf(fout, "<td bgcolor=#e0e0ff align=left><nobr>%s</nobr></td>", p->desc);
      fprintf(fout, "</tr>\n");
      row++;
      if (row==27) {
        fprintf(fout, "</table></td>\n");
        row=0;
      }
    }
  }
  fprintf(fout, "</tr></table>\n");
}

/* ---------------------------------------------------------------------- */
/* user interface */

void usage(FILE *fout) {
  fprintf(fout, "Usage: "NAME" [options]\n");
  fprintf(fout, "Options:\n");
  fprintf(fout, " -h, --help                 -- print usage info and exit\n");
  fprintf(fout, " -v, --version              -- print version info and exit\n");
  fprintf(fout, " -c, --codetable            -- print table of transliteration codes\n");
  fprintf(fout, " -u, --unicodetable n0 n1   -- print HTML table of unicodes from n0 to n1-1\n");
}

int main(int ac, char **av) {
  int n0, n1;

  ac--;
  av++;
  while (ac > 0) {
    if (strcmp(*av, "-h")==0 || strcmp(*av, "--help")==0) {
      fprintf(stdout, ""NAME" "VERSION": generate HTML from ASCII transliteration of Farsi\n");
      usage(stdout);
      exit(0);
    } else if (strcmp(*av, "-v")==0 || strcmp(*av, "--version")==0) {
      fprintf(stdout, ""NAME" "VERSION". "COPYRIGHT".\n");
      exit(0);
    } else if (strcmp(*av, "-c")==0 || strcmp(*av, "--codetable")==0) {
      codetable(stdout);
      exit(0);
    } else if (strcmp(*av, "-u")==0 || strcmp(*av, "--unicodetable")==0) {
      if (ac < 3) {
	fprintf(stderr, ""NAME": %s needs two arguments. See --help for more info.\n", *av);
	exit(1);
      }
      n0 = strtol(av[1], NULL, 0);
      n1 = strtol(av[2], NULL, 0);
      unicodetable(stdout, n0, n1);
      exit(0);
    } else if (**av == '-') {
      fprintf(stderr, ""NAME": unrecognized option -- %s. See --help for more info.\n", *av);
      exit(1);
    } else {
      fprintf(stderr, ""NAME": non-option argument not recognized. See --help for more info.\n");
      exit(1);
    }
  }
  translate(stdin, stdout);

  return 0;
}
